# awl.rankedmodel.models.py
from django.db import models, transaction

# ============================================================================

class RankedModel(models.Model):
    """Abstract model used to have all the inheritors ordered in the database
    by this model's ``rank`` field.   Ranks can either be across all instances
    of the inheriting model, or in a series of groups.  See
    :class:`RankedModel.group_filter` for details on implementing grouping.

    The ``rank`` field can be set and saved like any other field.  The
    overridden :class:`RankedModel.save` method maintains rank integrity.  The
    order is maintained but it is not guaranteed that there are not gaps in
    the rank count -- for simplicity re-ordering is not done on deletion.  If
    empty slots are a concern, use :class:`RankedModel.repack`.

    .. warning::

        Due to the use of the overridden ``save()`` for integrity caution must
        be employed when dealing with any ``update()`` calls or raw SQL as
        these will not call the ``save()`` method.

    Two admin helper functions are provided so you can do rank re-ording in
    the django admin.  To use the functions, add columns to your admin
    instances of the inheriting class::

        from awl.rankedmodel.admintools import (admin_link_move_up, 
            admin_link_move_down)

        @admin.register(Favourites)
        class FavouritesAdmin(admin.ModelAdmin):
            list_display = ('name', 'rank', 'move_up', 'move_down')

            def move_up(self, obj):
                return admin_link_move_up(obj)
            move_up.allow_tags = True
            move_up.short_description = 'Move Up Rank'

            def move_down(self, obj):
                return admin_link_move_down(obj)
            move_down.allow_tags = True
            move_down.short_description = 'Move Down Rank'

    :param rank:
        Ranked order of object
    """
    rank = models.PositiveSmallIntegerField(db_index=True)

    class Meta:
        abstract = True
        ordering = ['rank']

    def __init__(self, *args, **kwargs):
        super(RankedModel, self).__init__(*args, **kwargs)
        self._rank_at_load = self.rank

    def _process_new_rank_obj(self):
        # no id yet, this is the first time this object has been saved
        rank = getattr(self, 'rank', None)
        items = self.grouped_filter().order_by(
            'rank').select_for_update()
        count = items.count()
        if count == 0:
            self.rank = 1
        else:
            if not rank or rank > count + 1:
                # rank not set yet, or was set larger than largest item
                self.rank = count + 1
            elif rank < 1:
                self.rank = 1

            # count >= 1, need to move everybody else

            # rank was set to a specific value, need to re-order
            # everything that comes after it in the list
            for item in items[self.rank - 1:]:
                item.rank +=1
                item._rank_at_load = item.rank
                item.save(rerank=False)

        self._rank_at_load = self.rank

    def _process_moved_rank_obj(self):
        # rank changed, re-order it
        items = self.grouped_filter().order_by('rank').select_for_update()
        count = items.count()

        # check bounds on new rank
        if self.rank < 1:
            self.rank = 1
        elif self.rank > count:
            self.rank = count

        if self.rank < self._rank_at_load:
            # rank moved down
            start = self.rank - 1
            end = self._rank_at_load - 1
            delta = 1
        else:
            # rank moved up
            start = self._rank_at_load
            end = self.rank
            delta = -1

        for item in items[start:end]:
            item.rank += delta
            item._rank_at_load = item.rank
            item.save(rerank=False)

    @transaction.atomic
    def save(self, *args, **kwargs):
        """Overridden method that handles that re-ranking of objects and the
        integrity of the ``rank`` field.

        :param rerank:
            Added parameter, if True will rerank other objects based on the
            change in this save.  Defaults to True.  
        """
        rerank = kwargs.pop('rerank', True)
        if rerank:
            if not self.id:
                self._process_new_rank_obj()
            elif self.rank == self._rank_at_load:
                # nothing changed
                pass
            else:
                self._process_moved_rank_obj()

        super(RankedModel, self).save(*args, **kwargs)

    def grouped_filter(self):
        """This method should be overridden in order to allow groupings of
        ``RankModel`` objects.  The default is there is a single group which
        are all instances of the inheriting class.  

        An example with a grouped model would be::

            class Grouped(RankedModel):
                group_number = models.IntegerField()

                def grouped_filter(self):
                    return Grouped.objects.filter(
                        group_number=self.group_number)

        :returns:
            :class:`QuerySet` of ``RankedModel`` objects that are in the same
            group.
        """
        return self.__class__.objects.all()

    def repack(self):
        """Removes any blank ranks in the order."""
        items = self.grouped_filter().order_by('rank').select_for_update()
        for count, item in enumerate(items):
            item.rank = count + 1
            item.save(rerank=False)



from django.contrib import admin

from awl.rankedmodel import urls as ranked_urls
from awl.tests import views

from django.urls import include, path

urlpatterns = [
    path('admin/', admin.site.urls),

    path('rankedmodel/', include(ranked_urls)),

    path('awl_test_views/test_view_for_messages/', 
        views.test_view_for_messages, ),
]



# awl.tests.test_decorators.py
import json

from django.test import TestCase
from django.http import Http404

from awl.decorators import post_required, json_post_required
from awl.waelsteng import FakeRequest
from logthing.utils import silence_logging

# ============================================================================

class Worked(Exception):
    pass


@post_required
def post_view1(request):
    raise Worked()


@post_required(['foo'])
def post_view2(request):
    raise Worked()


@json_post_required('json_data', 'deserialized')
def json_post_view(request):
    return request.deserialized


@json_post_required('json_data')
def json_post_view2(request):
    return request.json_data

# ============================================================================

class DecoratorTest(TestCase):
    @silence_logging
    def test_post_required(self):
        request = FakeRequest()
        with self.assertRaises(Http404):
            post_view1(request)

        request = FakeRequest(method='POST')
        with self.assertRaises(Worked):
            post_view1(request)

        request = FakeRequest(method='POST', data={ 'foo':'bar', })
        with self.assertRaises(Worked):
            post_view2(request)

        request.POST = {
            'bar':'bar',
        }
        with self.assertRaises(Http404):
            post_view2(request)

    @silence_logging
    def test_json_post_required(self):
        # check that it ensures POST
        request = FakeRequest()
        with self.assertRaises(Http404):
            json_post_view(request)

        # check that it ensures post data
        request = FakeRequest(method='POST')
        with self.assertRaises(Http404):
            json_post_view(request)

        # good case
        content = { 'foo':'bar', }
        data = {
            'json_data':json.dumps(content),
        }
        request = FakeRequest(method='POST', data=data)
        result = json_post_view(request)
        self.assertEqual(content, result)

        # good case with optional request_name
        result = json_post_view2(request)
        self.assertEqual(content, result)



# awl.rankedmodel.admin.py

from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

# =============================================================================
# RankedModel Helper Methods
# =============================================================================

def admin_link_move_up(obj, link_text='↑'):
    """Returns a link to a view that moves the passed in object up in rank.

    :param obj:
        Object to move
    :param link_text:
        Text to display in the link.  Defaults to "↑"
    :returns:
        HTML link code to view for moving the object
    """
    if obj.rank == 1:
        return ''

    content_type = ContentType.objects.get_for_model(obj)
    link = reverse('awl-rankedmodel-move', args=(content_type.id, obj.id, 
        obj.rank - 1))

    return format_html('<a href="{}">{}</a>', link, link_text)


def admin_link_move_down(obj, link_text='↓'):
    """Returns a link to a view that moves the passed in object down in rank.

    :param obj:
        Object to move
    :param link_text:
        Text to display in the link.  Defaults to "↓"
    :returns:
        HTML link code to view for moving the object
    """
    if obj.rank == obj.grouped_filter().count():
        return ''

    content_type = ContentType.objects.get_for_model(obj)
    link = reverse('awl-rankedmodel-move', args=(content_type.id, obj.id, 
        obj.rank + 1))

    return format_html('<a href="{}">{}</a>', link, link_text)


def admin_move_links(obj, up_text='↑', down_text='↓'):
    """Returns a formatted column with up to two links in it for moving the
    passed in object up or down in rank.

    :param obj:
        Object to move
    :param up_text:
        Text to display in the up link.  Defaults to "↑"
    :param down_text:
        Text to display in down the link.  Defaults to "↓"
    :returns:
        HTML link code to view for moving the object
    """
    show_up = True
    show_down = True

    if obj.rank == 1:
        show_up = False

    if obj.rank == obj.grouped_filter().count():
        show_down = False

    html = f'<span style="width:{len(up_text)+1}ex; display:inline-block">'
    if show_up:
        content_type = ContentType.objects.get_for_model(obj)
        link = reverse('awl-rankedmodel-move', args=(content_type.id, obj.id, 
            obj.rank - 1))
        html += f'<a href="{link}">{up_text}</a>'
    else:
        html += '&nbsp;'

    html += '</span>&nbsp;' + \
        f'<span style="width:{len(down_text)+1}ex; display:inline-block">'
    if show_down:
        content_type = ContentType.objects.get_for_model(obj)
        link = reverse('awl-rankedmodel-move', 
            args=(content_type.id, obj.id, obj.rank + 1))
        html += f'<a href="{link}">{down_text}</a>'

    html += '</span>'

    return mark_safe(html)



# awl.tests.test_tags.py

from django.template import Context, Template, TemplateSyntaxError
from django.test import TestCase

# ============================================================================

class TagTests(TestCase):
    def test_getitem(self):
        t = """
        {% load awltags %}
        
        {{d|getitem:key}}"""

        template = Template(t)
        data = {
            'd': {
                'foo':'bar',
            },
            'key':'foo',
        }
        context = Context(data)
        result = template.render(context)
        self.assertEqual('bar', result.strip())

        # -- test bad key
        data['key'] = 'missing'
        context = Context(data)
        result = template.render(context)
        self.assertEqual('', result.strip())

    def test_accessor(self):
        class Dummy(object):
            pass

        d = Dummy()
        d.one = Dummy()
        d.one.two = {
            'three': {
                'four': 'five',
            }
        }

        context_data = {
            'myobj':d,
            'attr1':'one',
            'key1':'three',
        }

        t = """
        {% load awltags %}
        
        {% accessor myobj attr1 'two' [key1] ['four'] %}
        """

        template = Template(t)
        context = Context(context_data)
        result = template.render(context)
        self.assertEqual('five', result.strip())

        # -- test "as" functionality
        t = """
        {% load awltags %}
        
        {% accessor myobj attr1 'two' [key1] ['four'] as thing %}
        """
        template = Template(t)
        context = Context(context_data)
        result = template.render(context)
        self.assertEqual('', result.strip())
        self.assertEqual('five', context['thing'])

        # -- test failure modes

        # test no attributes
        with self.assertRaises(TemplateSyntaxError):
            t = """
            {% load awltags %}
            
            {% accessor %}
            """

            template = Template(t)

        # test 1 attribute
        with self.assertRaises(TemplateSyntaxError):
            t = """
            {% load awltags %}
            
            {% accessor myobj %}
            """
            template = Template(t)

        # test KeyError results in blank 
        t = """
        {% load awltags %}
        
        {% accessor myobj attr1 'foo' [key1] ['four'] %}
        """
        template = Template(t)
        context = Context(context_data)
        result = template.render(context)
        self.assertEqual('', result.strip())

        # test KeyError results blank context variable
        t = """
        {% load awltags %}
        
        {% accessor myobj attr1 'foo' [key1] ['four'] as thing %}
        """
        template = Template(t)
        context = Context(context_data)
        result = template.render(context)
        self.assertEqual('', result.strip())
        self.assertEqual('', context['thing'])

    def test_nop(self):
        t = "{% load awltags %}{% nop 'thing and stuff' %}"
        
        template = Template(t)
        context = Context({})
        result = template.render(context)
        self.assertEqual('', result)

    def test_jsonify(self):
        t = "{% load awltags %}let actors = {% jsonify actors %};"

        actors = {
            'Pitt':'Brad',
        }
        
        template = Template(t)
        context = Context({'actors':actors})
        result = template.render(context)

        expected = """let actors = {"Pitt": "Brad"};"""
        self.assertEqual(expected, result)



# awl.tests.test_models.py
from django.test import TestCase

from awl.models import Counter, Lock, Choices, QuerySetChain
from awl.utils import refetch

# ============================================================================

class ModelsTest(TestCase):
    def test_counter(self):
        count = Counter.objects.create(name='foo')
        Counter.increment('foo')
        count = refetch(count)
        self.assertEqual(1, count.value)

    def test_lock(self):
        # not much to test here except that it doesn't blow up
        Lock.objects.create(name='foo')
        Lock.lock_until_commit('foo')

    def test_choices(self):
        class Colours(Choices):
            RED = 'r'
            LIGHT_ORANGE = 'o'
            BLUE = ('b', 'Blueish')

        class MoreColours(Colours):
            GREEN = 'g'

        # check the created values
        self.assertEqual(MoreColours.RED, 'r')
        self.assertEqual(MoreColours.LIGHT_ORANGE, 'o')
        self.assertEqual(MoreColours.BLUE, 'b')
        self.assertEqual(MoreColours.GREEN, 'g')

        # check conversion to django-style choices list
        l = list(MoreColours)
        self.assertEqual(len(l), 4)
        self.assertIn(('r', 'Red'), l)
        self.assertIn(('b', 'Blueish'), l)
        self.assertIn(('g', 'Green'), l)
        self.assertIn(('o', 'Light Orange'), l)

        # check gettr method
        self.assertEqual('Red', Colours.get_value('r'))
        self.assertEqual('Red', MoreColours.get_value('r'))

    def test_queryset_chain(self):
        # create some object to query
        from django.contrib.auth.models import User, Group
        User.objects.create(username='u1')
        User.objects.create(username='u2')
        User.objects.create(username='u3')
        Group.objects.create(name='g1')
        Group.objects.create(name='g2')

        # get 3 users and two groups
        users = User.objects.all()
        groups = Group.objects.all()

        # base case for chaining
        chain = QuerySetChain(users, groups)
        self.assertEqual(5, chain.count())

        # test slicing
        for item in chain[:3]:
            self.assertIsInstance(item, User)

        self.assertEqual(groups[0], chain[3])

        # trigger internal _clone(), make sure it doesn't blow up
        chain._clone()



# awl.tests.test_admintools.py
from django.contrib.admin.utils import label_for_field
from django.test import TestCase
from django.utils.safestring import SafeData

from screwdriver import parse_link

from awl.waelsteng import AdminToolsMixin
from awl.tests.models import (Author, Book, Chapter, Driver, VehicleMake, 
    VehicleModel)
from awl.tests.admin import (BookAdmin, ChapterAdmin, DriverAdmin,
    VehicleModelAdmin)

# ============================================================================

class AdminToolsTest(TestCase, AdminToolsMixin):
    def test_admin_obj_mixin(self):
        # setup the admin site and id
        self.initiate()

        book_admin = BookAdmin(Book, self.site)
        chapter_admin = ChapterAdmin(Chapter, self.site)

        tolstoy = Author.objects.create(name='Tolstoy')
        war = Book.objects.create(name='War and Peace', author=tolstoy)
        ch1 = Chapter.objects.create(name='Part I.I', book=war)

        # make sure we get a safe string
        html = self.field_value(book_admin, war, 'show_author')
        self.assertTrue(issubclass(html.__class__, SafeData))

        # check the basic __str__ named link from Book to Author
        html = self.field_value(book_admin, war, 'show_author')
        url, text = parse_link(html)
        self.assertEqual('Author(id=1 Tolstoy)', text)
        self.assertEqual('/admin/tests/author/?id__exact=1', url)

        # check the template based name link of Book from Chapter
        html = self.field_value(chapter_admin, ch1, 'show_book')
        url, text = parse_link(html)
        self.assertEqual('Book.id=1', text)
        self.assertEqual('/admin/tests/book/?id__exact=1', url)

        # check the double dereferenced Author from Chapter
        html = self.field_value(chapter_admin, ch1, 'show_author')
        url, text = parse_link(html)
        self.assertEqual('Author(id=1 Tolstoy)', text)
        self.assertEqual('/admin/tests/author/?id__exact=1', url)

        # check readonly, double dereferenced Author from Chapter
        result = self.field_value(chapter_admin, ch1, 'readonly_author')
        self.assertEqual('Author(id=1 Tolstoy)', result)
        result = self.field_value(chapter_admin, ch1, 'readonly_book')
        self.assertEqual('RO Book.id=1', result)

        # check the title got set correctly
        label = label_for_field('show_book', ch1, chapter_admin)
        self.assertEqual(label, 'My Book')

        # check that empty values work properly
        ch2 = Chapter.objects.create(name='Part I.II')
        result = self.field_value(chapter_admin, ch2, 'show_author')
        self.assertEqual('', result)

        book = Book.objects.create(name='book')
        ch2.book = book
        ch2.save()
        result = self.field_value(chapter_admin, ch2, 'show_author')
        self.assertEqual('', result)

        result = self.field_value(chapter_admin, ch2, 'readonly_author')
        self.assertEqual('', result)


    def test_fancy_modeladmin(self):
        # setup the admin site and id
        self.initiate()

        driver_admin = DriverAdmin(Driver, self.site)
        vehiclemodel_admin = VehicleModelAdmin(VehicleModel, self.site)

        toyota = VehicleMake.objects.create(name='Toyota')
        tercel = VehicleModel.objects.create(name='Tercel', vehiclemake=toyota,
            year=1999)
        bob = Driver.objects.create(name='Bob', vehiclemodel=tercel, 
            rating=0.35)

        vehiclemake_field = driver_admin.list_display[2]
        vehiclemodel_field = driver_admin.list_display[3]
        ro_vehiclemake_field = driver_admin.list_display[4]
        ro_vehiclemodel_field = driver_admin.list_display[5]
        rating_field = driver_admin.list_display[6]
        rating_field_templated = driver_admin.list_display[7]

        year_field = vehiclemodel_admin.list_display[2]
        default_set_field = vehiclemodel_admin.list_display[3]
        custom_set_field = vehiclemodel_admin.list_display[4]

        # make sure we get a safe string
        html = self.field_value(driver_admin, bob, vehiclemake_field)
        self.assertTrue(issubclass(html.__class__, SafeData))

        # check the basic __str__ named link from Student to Teacher
        html = self.field_value(driver_admin, bob, vehiclemake_field)
        url, text = parse_link(html)
        self.assertEqual('VehicleMake(id=1 Toyota)', text)
        self.assertEqual('/admin/tests/vehiclemake/?id__exact=1', url)

        # check the template based name link of Course from Student
        html = self.field_value(driver_admin, bob, vehiclemodel_field)
        url, text = parse_link(html)
        self.assertEqual('Toyota Tercel id=1', text)
        self.assertEqual('/admin/tests/vehiclemodel/?id__exact=1', url)

        # check readonly fields
        result = self.field_value(driver_admin, bob, ro_vehiclemake_field)
        self.assertEqual('VehicleMake(id=1 Toyota)', result)
        result = self.field_value(driver_admin, bob, ro_vehiclemodel_field)
        self.assertEqual('RO Toyota Tercel id=1', result)

        # check the titles got set correctly
        label = label_for_field(vehiclemodel_field, bob, driver_admin)
        self.assertEqual(label, 'My Vehicle Model')
        label = label_for_field(ro_vehiclemodel_field, bob, driver_admin)
        self.assertEqual(label, 'RO Vehicle Model')

        # check that empty values work properly
        fred = Driver.objects.create(name='Fred')
        result = self.field_value(driver_admin, fred, vehiclemodel_field)
        self.assertEqual('<i>no model</i>', result)
        result = self.field_value(driver_admin, fred, vehiclemake_field)
        self.assertEqual('', result)

        result = self.field_value(driver_admin, fred, ro_vehiclemodel_field)
        self.assertEqual('', result)
        result = self.field_value(driver_admin, fred, ro_vehiclemake_field)
        self.assertEqual('', result)

        # check formatted field
        result = self.field_value(driver_admin, bob, rating_field)
        self.assertEqual('0.3', result)

        # check templated field
        result = self.field_value(driver_admin, bob, rating_field_templated)
        self.assertEqual('Bob 0.35', result)

        # check display with title
        result = self.field_value(vehiclemodel_admin, tercel, year_field)
        self.assertEqual(1999, result)

        label = label_for_field(year_field, tercel, vehiclemodel_admin)
        self.assertEqual(label, 'YEAR TITLE')

        # check display with title with empty result
        yearless = VehicleModel.objects.create(name='Yearless', 
            vehiclemake=toyota)
        result = self.field_value(vehiclemodel_admin, yearless, year_field)
        self.assertEqual('<i>no year</i>', result)

        # Validate the add_fk_link method with default values
        html = self.field_value(vehiclemodel_admin, tercel,
            default_set_field)
        url, text = parse_link(html)
        self.assertEqual('1 Driver', text)
        self.assertEqual('/admin/tests/driver/?vehiclemodel__id=1', url)

        label = label_for_field(default_set_field, tercel, vehiclemodel_admin)
        self.assertEqual('Driver', label)

        # Validate the add_fk_link method with custom values
        html = self.field_value(vehiclemodel_admin, tercel,
            custom_set_field)
        url, text = parse_link(html)
        self.assertEqual('Tercel 1 Driver Title', text)
        self.assertEqual('/admin/tests/driver/?vehiclemodel__id=1', url)

        label = label_for_field(custom_set_field, tercel, vehiclemodel_admin)
        self.assertEqual('Driver Title', label)

        # Validate add_fk_link, vehicle with no drivers
        driverless = VehicleModel.objects.create(name='Driverless', 
            vehiclemake=toyota)
        html = self.field_value(vehiclemodel_admin, driverless,
            custom_set_field)
        self.assertEqual('<i>no drivers</i>', html)
